import logging
from typing import Any, Dict, List, Optional
from uuid import UUID

import requests

from starpoint._utils import (
    _build_header,
    _check_collection_identifier_collision,
    _validate_host,
)

LOGGER = logging.getLogger(__name__)

# Host
WRITER_URL = "https://writer.starpoint.ai"

# Endpoints
COLLECTIONS_PATH = "/api/v1/collections"
DOCUMENTS_PATH = "/api/v1/documents"

# Error and warning messages
DIMENSIONALITY_ERROR = "Dimensionality must be greater than 0."
EMBEDDING_METADATA_LENGTH_MISMATCH_WARNING = (
    "The length of the embeddings and document_metadata provided are different. There may be a mismatch "
    "between embeddings and the expected document metadata length; this may cause undesired collection insert or update."
)
SSL_ERROR_MSG = "Request failed due to SSLError. Error is likely due to invalid API key. Please check if your API is correct and still valid."


class Writer(object):
    """Client for the Writer endpoints. If you do not need to separate reading or
    writing for, consider using the [`Client` object](#client-objects)."""

    def __init__(self, api_key: UUID, host: Optional[str] = None):
        if host is None:
            host = WRITER_URL

        self.host = _validate_host(host)
        self.api_key = api_key

    def delete(
        self,
        documents: List[str],
        collection_id: Optional[str] = None,
        collection_name: Optional[str] = None,
    ) -> Dict[Any, Any]:
        """Remove documents in an existing collection.

        Args:
            documents: The documents to remove from the collection.
            collection_id: The collection's id to remove the documents from.
                This or the `collection_name` needs to be provided.
            collection_name: The collection's name to remove the documents from.
                This or the `collection_id` needs to be provided.

        Returns:
            dict: delete response json

        Raises:
            ValueError: If neither collection id and collection name are provided.
            ValueError: If both collection id and collection name are provided.
        """
        _check_collection_identifier_collision(collection_id, collection_name)
        # TODO: Be safe and make sure the item passed through that doesn't hold a value is a None

        # TODO: filter through documents to make sure values are what we expect them to be
        """
        dict(
            collection_id="collection_id_example",
            collection_name="collection_name_example",
            documents=["uuid-uuid-uuid"]
            ],
        )
        """

        request_data = dict(
            collection_id=collection_id,
            collection_name=collection_name,
            documents=[str(document) for document in documents],
        )
        try:
            response = requests.delete(
                url=f"{self.host}{DOCUMENTS_PATH}",
                json=request_data,
                headers=_build_header(
                    api_key=self.api_key,
                    additional_headers={"Content-Type": "application/json"},
                ),
            )
        except requests.exceptions.SSLError as e:
            LOGGER.error(SSL_ERROR_MSG)
            raise e

        if not response.ok:
            LOGGER.error(
                f"Request failed with status code {response.status_code} "
                f"and the following message:\n{response.text}"
            )
            return {}
        return response.json()

    def column_delete(
        self,
        embeddings: List[List[float]],
        document_metadatas: List[Dict[Any, Any]],
        collection_id: Optional[str] = None,
        collection_name: Optional[str] = None,
    ) -> Dict[Any, Any]:
        """Deletes documents from an existing collection by embedding and document metadata arrays.
        The arrays are zipped together and updates the document in the order of the two arrays.

        Args:
            embeddings: A list of embeddings.
                Order of the embeddings should match the document_metadatas.
            document_metadatas: A list of metadata to be associated with embeddings.
                Order of these metadatas should match the embeddings.
            collection_id: The collection's id where the documents will be deleted.
                This or the `collection_name` needs to be provided.
            collection_name: The collection's name where the documents will be deleted.
                This or the `collection_id` needs to be provided.

        Returns:
            dict: delete response json

        Raises:
            ValueError: If neither collection id and collection name are provided.
            ValueError: If both collection id and collection name are provided.
            requests.exceptions.SSLError: Failure likely due to network issues.
        """
        if len(embeddings) != len(document_metadatas):
            LOGGER.warning(EMBEDDING_METADATA_LENGTH_MISMATCH_WARNING)

        documents = [
            {
                "embedding": embedding,
                "metadata": document_metadata,
            }
            for embedding, document_metadata in zip(embeddings, document_metadatas)
        ]

        return self.delete(
            documents=documents,
            collection_id=collection_id,
            collection_name=collection_name,
        )

    def insert(
        self,
        documents: List[Dict[Any, Any]],
        collection_id: Optional[str] = None,
        collection_name: Optional[str] = None,
    ) -> Dict[Any, Any]:
        """Insert documents into an existing collection.

        Args:
            documents: The documents to insert into the collection.
            collection_id: The collection's id to insert the documents to.
                This or the `collection_name` needs to be provided.
            collection_name: The collection's name to insert the documents to.
                This or the `collection_id` needs to be provided.

        Returns:
            dict: insert response json

        Raises:
            ValueError: If neither collection id and collection name are provided.
            ValueError: If both collection id and collection name are provided.
            requests.exceptions.SSLError: Failure likely due to network issues.
        """

        _check_collection_identifier_collision(collection_id, collection_name)
        # TODO: Be safe and make sure the item passed through that doesn't hold a value is a None

        # TODO: filter through documents to make sure values are what we expect them to be
        """
        dict(
            collection_id="collection_id_example",
            collection_name="collection_name_example",
            documents=[
                dict(
                    embedding=[3.14],
                    metadata=dict(),
                )
            ],
        )
        """

        request_data = dict(
            collection_id=collection_id,
            collection_name=collection_name,
            documents=documents,
        )
        try:
            response = requests.post(
                url=f"{self.host}{DOCUMENTS_PATH}",
                json=request_data,
                headers=_build_header(
                    api_key=self.api_key,
                    additional_headers={"Content-Type": "application/json"},
                ),
            )
        except requests.exceptions.SSLError as e:
            LOGGER.error(SSL_ERROR_MSG)
            raise e

        if not response.ok:
            LOGGER.error(
                f"Request failed with status code {response.status_code} "
                f"and the following message:\n{response.text}"
            )
            return {}
        return response.json()

    def column_insert(
        self,
        embeddings: List[List[float]],
        document_metadatas: List[Dict[Any, Any]],
        collection_id: Optional[str] = None,
        collection_name: Optional[str] = None,
    ) -> Dict[Any, Any]:
        """Insert documents into an existing collection by embedding and document metadata arrays.
        The arrays are zipped together and inserted as a document in the order of the two arrays.

        Args:
            embeddings: A list of embeddings.
                Order of the embeddings should match the document_metadatas.
            document_metadatas: A list of metadata to be associated with embeddings.
                Order of these metadatas should match the embeddings.
            collection_id: The collection's id to insert the documents to.
                This or the `collection_name` needs to be provided.
            collection_name: The collection's name to insert the documents to.
                This or the `collection_id` needs to be provided.

        Returns:
            dict: insert response json

        Raises:
            ValueError: If neither collection id and collection name are provided.
            ValueError: If both collection id and collection name are provided.
            requests.exceptions.SSLError: Failure likely due to network issues.
        """
        if len(embeddings) != len(document_metadatas):
            LOGGER.warning(EMBEDDING_METADATA_LENGTH_MISMATCH_WARNING)

        documents = [
            {
                "embedding": embedding,
                "metadata": document_metadata,
            }
            for embedding, document_metadata in zip(embeddings, document_metadatas)
        ]

        return self.insert(
            documents=documents,
            collection_id=collection_id,
            collection_name=collection_name,
        )

    def update(
        self,
        documents: List[Dict[Any, Any]],
        collection_id: Optional[str] = None,
        collection_name: Optional[str] = None,
    ) -> Dict[Any, Any]:
        """Update documents in an existing collection.

        Args:
            documents: The documents to update in the collection.
            collection_id: The collection's id where the documents will be updated.
                This or the `collection_name` needs to be provided.
            collection_name: The collection's name where the documents will be updated.
                This or the `collection_id` needs to be provided.

        Returns:
            dict: update response json

        Raises:
            ValueError: If neither collection id and collection name are provided.
            ValueError: If both collection id and collection name are provided.
            requests.exceptions.SSLError: Failure likely due to network issues.
        """
        _check_collection_identifier_collision(collection_id, collection_name)
        # TODO: Be safe and make sure the item passed through that doesn't hold a value is a None

        # TODO: filter through documents to make sure values are what we expect them to be
        """
        dict(
            collection_id="collection_id_example",
            collection_name="collection_name_example",
            documents=[
                dict(
                    id:"UUID-UUID-UUID",
                    metadata={},
                )
            ],
        )
        """

        request_data = dict(
            collection_id=collection_id,
            collection_name=collection_name,
            documents=documents,
        )
        try:
            response = requests.patch(
                url=f"{self.host}{DOCUMENTS_PATH}",
                json=request_data,
                headers=_build_header(
                    api_key=self.api_key,
                    additional_headers={"Content-Type": "application/json"},
                ),
            )
        except requests.exceptions.SSLError as e:
            LOGGER.error(SSL_ERROR_MSG)
            raise e

        if not response.ok:
            LOGGER.error(
                f"Request failed with status code {response.status_code} "
                f"and the following message:\n{response.text}"
            )
            return {}
        return response.json()

    def column_update(
        self,
        embeddings: List[List[float]],
        document_metadatas: List[Dict[Any, Any]],
        collection_id: Optional[str] = None,
        collection_name: Optional[str] = None,
    ) -> Dict[Any, Any]:
        """Updates documents for an existing collection by embedding and document metadata arrays.
        The arrays are zipped together and updates the document in the order of the two arrays.

        Args:
            embeddings: A list of embeddings.
                Order of the embeddings should match the document_metadatas.
            document_metadatas: A list of metadata to be associated with embeddings.
                Order of these metadatas should match the embeddings.
            collection_id: The collection's id where the documents will be updated.
                This or the `collection_name` needs to be provided.
            collection_name: The collection's name where the documents will be updated.
                This or the `collection_id` needs to be provided.

        Returns:
            dict: update response json

        Raises:
            ValueError: If neither collection id and collection name are provided.
            ValueError: If both collection id and collection name are provided.
            requests.exceptions.SSLError: Failure likely due to network issues.
        """
        if len(embeddings) != len(document_metadatas):
            LOGGER.warning(EMBEDDING_METADATA_LENGTH_MISMATCH_WARNING)

        documents = [
            {
                "embedding": embedding,
                "metadata": document_metadata,
            }
            for embedding, document_metadata in zip(embeddings, document_metadatas)
        ]

        return self.update(
            documents=documents,
            collection_id=collection_id,
            collection_name=collection_name,
        )

    def create_collection(
        self, collection_name: str, dimensionality: int
    ) -> Dict[Any, Any]:
        """Creates a collection by name and dimensionality. Dimensionality
        should be greater than 0.

        Args:
            collection_name: The name of the collection that will be created.
            dimensionality: The number of dimensions the collection will have.
                Must be an int larger than 0.

        Returns:
            dict: create collections response json

        Raises:
            ValueError: If dimensionality is 0 or less.
            requests.exceptions.SSLError: Failure likely due to network issues.
        """

        if dimensionality <= 0:
            raise ValueError(DIMENSIONALITY_ERROR)

        request_data = dict(
            name=collection_name,
            dimensionality=dimensionality,
        )
        try:
            response = requests.post(
                url=f"{self.host}{COLLECTIONS_PATH}",
                json=request_data,
                headers=_build_header(
                    api_key=self.api_key,
                    additional_headers={"Content-Type": "application/json"},
                ),
            )
        except requests.exceptions.SSLError as e:
            LOGGER.error(SSL_ERROR_MSG)
            raise e

        if not response.ok:
            LOGGER.error(
                f"Request failed with status code {response.status_code} "
                f"and the following message:\n{response.text}"
            )
            return {}
        return response.json()

    def delete_collection(self, collection_id: str) -> Dict[Any, Any]:
        """Deletes a collection.

        Args:
            collection_id: The id of the collection that will be deleted.

        Returns:
            dict: deleted collection response json

        Raises:
            requests.exceptions.SSLError: Failure likely due to network issues.
        """
        request_data = dict(
            collection_id=collection_id,
        )
        try:
            response = requests.delete(
                url=f"{self.host}{COLLECTIONS_PATH}",
                json=request_data,
                headers=_build_header(
                    api_key=self.api_key,
                    additional_headers={"Content-Type": "application/json"},
                ),
            )
        except requests.exceptions.SSLError as e:
            LOGGER.error(SSL_ERROR_MSG)
            raise e

        if not response.ok:
            LOGGER.error(
                f"Request failed with status code {response.status_code} "
                f"and the following message:\n{response.text}"
            )
            return {}
        return response.json()



from typing import List
from unittest.mock import MagicMock, patch
from uuid import UUID, uuid4

import pytest
from _pytest.monkeypatch import MonkeyPatch
from requests.exceptions import SSLError

from starpoint import embedding


@pytest.fixture(scope="session")
def api_uuid() -> UUID:
    return uuid4()


@pytest.fixture(scope="session")
@patch("starpoint._utils._check_host_health")
def mock_embedding_client(
    host_health_mock: MagicMock, api_uuid: UUID
) -> embedding.EmbeddingClient:
    return embedding.EmbeddingClient(api_uuid)


def test_embedding_default_init(
    mock_embedding_client: embedding.EmbeddingClient, api_uuid: UUID
):
    assert mock_embedding_client.host
    assert mock_embedding_client.host == embedding.EMBEDDING_URL
    assert mock_embedding_client.api_key == api_uuid


@patch("starpoint.embedding._validate_host")
def test_embedding_init_non_default_host(
    mock_host_validator: MagicMock, api_uuid: UUID
):
    test_host = "http://www.example.com"
    test_embedding_client = embedding.EmbeddingClient(api_key=api_uuid, host=test_host)

    mock_host_validator.assert_called_once_with(test_host)
    # This assert needs to be after assert_called_once_with to make sure it doesn't confound the result
    assert test_embedding_client.host == mock_host_validator()
    assert test_embedding_client.api_key == api_uuid


@patch("starpoint.embedding.requests")
def test_embedding_embed_not_200(
    requests_mock: MagicMock,
    mock_embedding_client: embedding.EmbeddingClient,
    monkeypatch: MonkeyPatch,
):
    requests_mock.post().ok = False

    expected_json = {}

    logger_mock = MagicMock()
    monkeypatch.setattr(embedding, "LOGGER", logger_mock)

    actual_json = mock_embedding_client.embed(["asdf"], embedding.EmbeddingModel.MINILM)

    requests_mock.post.assert_called()
    logger_mock.error.assert_called_once()
    assert actual_json == expected_json


@patch("starpoint.embedding.requests")
def test_embedding_embed_SSLError(
    requests_mock: MagicMock,
    mock_embedding_client: embedding.EmbeddingClient,
    monkeypatch: MonkeyPatch,
):
    requests_mock.exceptions.SSLError = SSLError
    requests_mock.post.side_effect = SSLError("mock exception")

    logger_mock = MagicMock()
    monkeypatch.setattr(embedding, "LOGGER", logger_mock)

    with pytest.raises(SSLError, match="mock exception"):
        mock_embedding_client.embed(["asdf"], embedding.EmbeddingModel.MINILM)

    logger_mock.error.assert_called_once_with(embedding.SSL_ERROR_MSG)


@patch("starpoint.embedding.EmbeddingClient.embed_items")
@patch("starpoint.embedding.requests")
def test_embedding_embed(
    requests_mock: MagicMock,
    embed_items_mock: MagicMock,
    mock_embedding_client: embedding.EmbeddingClient,
):
    input_text = "asdf"
    input_model = embedding.EmbeddingModel.MINILM
    expected_item = [{"text": input_text, "metadata": None}]

    actual_json = mock_embedding_client.embed([input_text], input_model)

    embed_items_mock.assert_called_once_with(
        text_embedding_items=expected_item, model=input_model
    )


@patch("starpoint.embedding.EmbeddingClient.embed_items")
@patch("starpoint.embedding.requests")
def test_embedding_embed_and_join_metadata_by_columns(
    requests_mock: MagicMock,
    embed_items_mock: MagicMock,
    mock_embedding_client: embedding.EmbeddingClient,
):
    input_text = "asdf"
    input_metadata = {"label": "asdf"}
    input_model = embedding.EmbeddingModel.MINILM
    expected_item = [{"text": input_text, "metadata": input_metadata}]

    actual_json = mock_embedding_client.embed_and_join_metadata_by_columns(
        [input_text], [input_metadata], input_model
    )

    embed_items_mock.assert_called_once_with(
        text_embedding_items=expected_item, model=input_model
    )


@patch("starpoint.embedding.EmbeddingClient.embed_items")
@patch("starpoint.embedding.requests")
def test_embedding_embed_and_join_metadata_by_columns_non_list_texts(
    requests_mock: MagicMock,
    embed_items_mock: MagicMock,
    mock_embedding_client: embedding.EmbeddingClient,
):
    input_metadata = {"label": "asdf"}
    input_model = embedding.EmbeddingModel.MINILM

    with pytest.raises(ValueError):
        mock_embedding_client.embed_and_join_metadata_by_columns(
            "not_list_texts", [input_metadata], input_model
        )


@patch("starpoint.embedding.EmbeddingClient.embed_items")
@patch("starpoint.embedding.requests")
def test_embedding_embed_and_join_metadata_by_columns_non_list_metadatas(
    requests_mock: MagicMock,
    embed_items_mock: MagicMock,
    mock_embedding_client: embedding.EmbeddingClient,
):
    input_text = "asdf"
    input_model = embedding.EmbeddingModel.MINILM

    with pytest.raises(ValueError):
        mock_embedding_client.embed_and_join_metadata_by_columns(
            [input_text], {"label": "not_list_metadatas"}, input_model
        )


@pytest.mark.parametrize(
    "input_text,input_metadata",
    [
        [
            ["embed_text1", "embed_text2"],
            [{"label": "label1"}],
        ],
        [
            ["embed_text1"],
            [{"label": "label1"}, {"label": "label2"}],
        ],
    ],
)
@patch("starpoint.embedding.EmbeddingClient.embed_items")
@patch("starpoint.embedding.requests")
def test_embedding_embed_and_join_metadata_by_columns_mismatch_list(
    requests_mock: MagicMock,
    embed_items_mock: MagicMock,
    mock_embedding_client: embedding.EmbeddingClient,
    input_text: List,
    input_metadata: List,
    monkeypatch: MonkeyPatch,
):
    input_model = embedding.EmbeddingModel.MINILM

    logger_mock = MagicMock()
    monkeypatch.setattr(embedding, "LOGGER", logger_mock)

    actual_json = mock_embedding_client.embed_and_join_metadata_by_columns(
        input_text, input_metadata, input_model
    )

    logger_mock.warning.assert_called_once_with(
        embedding.TEXT_METADATA_LENGTH_MISMATCH_WARNING
    )


@patch("starpoint.embedding.EmbeddingClient.embed_and_join_metadata_by_columns")
@patch("starpoint.embedding.requests")
def test_embedding_embed_and_join_metadata(
    requests_mock: MagicMock,
    embed_and_join_metadata_by_columns_mock: MagicMock,
    mock_embedding_client: embedding.EmbeddingClient,
):
    input_text = "embed text"
    input_metadata = {
        "metadata1": "metadata1",
        "metadata2": "metadata2",
    }
    embed_key = "text"
    input_dict = {embed_key: input_text}
    input_dict.update(input_metadata)
    test_embedding_items = [input_dict]

    expected_item = [{"text": input_text, "metadata": input_metadata}]
    input_model = embedding.EmbeddingModel.MINILM

    actual_json = mock_embedding_client.embed_and_join_metadata(
        test_embedding_items, embed_key, input_model
    )

    embed_and_join_metadata_by_columns_mock.assert_called_once_with(
        texts=[input_text], metadatas=[input_metadata], model=input_model
    )


@patch("starpoint.embedding.EmbeddingClient.embed_and_join_metadata_by_columns")
@patch("starpoint.embedding.requests")
def test_embedding_embed_and_join_metadata_no_embed_key(
    requests_mock: MagicMock,
    embed_and_join_metadata_by_columns_mock: MagicMock,
    mock_embedding_client: embedding.EmbeddingClient,
):
    input_text = "embed text"
    input_metadata = {
        "metadata1": "metadata1",
        "metadata2": "metadata2",
    }
    input_dict = {"text": input_text}
    input_dict.update(input_metadata)
    test_embedding_items = [input_dict]

    input_model = embedding.EmbeddingModel.MINILM
    embed_key = "no key"

    with pytest.raises(ValueError):
        actual_json = mock_embedding_client.embed_and_join_metadata(
            test_embedding_items, embed_key, input_model
        )


@patch("starpoint.embedding.EmbeddingClient.embed_and_join_metadata_by_columns")
@patch("starpoint.embedding.requests")
def test_embedding_embed_and_join_metadata_no_values(
    requests_mock: MagicMock,
    embed_and_join_metadata_by_columns_mock: MagicMock,
    mock_embedding_client: embedding.EmbeddingClient,
):
    input_model = embedding.EmbeddingModel.MINILM
    with pytest.raises(ValueError):
        actual_json = mock_embedding_client.embed_and_join_metadata(
            [], "text", input_model
        )


@patch("starpoint.embedding.requests")
def test_embedding_embed_items(
    requests_mock: MagicMock,
    mock_embedding_client: embedding.EmbeddingClient,
):
    requests_mock.post().ok = True
    test_value = {"mock_return": "value"}
    requests_mock.post().json.return_value = test_value

    expected_json = {}

    actual_json = mock_embedding_client.embed_items(
        [{"text": "asdf", "metadata": {"label": "asdf"}}],
        embedding.EmbeddingModel.MINILM,
    )

    requests_mock.post.assert_called()
    requests_mock.post().json.assert_called()
    assert actual_json == test_value



import logging
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Union
from uuid import UUID

import openai
import requests
import validators

from starpoint import reader, writer, _utils

LOGGER = logging.getLogger(__name__)


class Client(object):
    """Client that combines Reader and Writer. It is recommended that one use this client rather than
    Reader and Writer independently."""

    def __init__(
        self,
        api_key: UUID,
        reader_host: Optional[str] = None,
        writer_host: Optional[str] = None,
    ):
        self.writer = writer.Writer(api_key=api_key, host=writer_host)
        self.reader = reader.Reader(api_key=api_key, host=reader_host)

        # Consider a wrapper around openai once this class gets bloated
        self.openai = None

    def delete(
        self,
        documents: List[str],
        collection_id: Optional[str] = None,
        collection_name: Optional[str] = None,
    ) -> Dict[Any, Any]:
        """Remove documents in an existing collection. `delete()` method from [`Writer`](#writer-objects).

        Args:
            documents: The documents to remove from the collection.
            collection_id: The collection's id to remove the documents from.
                This or the `collection_name` needs to be provided.
            collection_name: The collection's name to remove the documents from.
                This or the `collection_id` needs to be provided.

        Returns:
            dict: delete response json

        Raises:
            ValueError: If neither collection id and collection name are provided.
            ValueError: If both collection id and collection name are provided.
        """
        return self.writer.delete(
            documents=documents,
            collection_id=collection_id,
            collection_name=collection_name,
        )

    def column_delete(
        self,
        embeddings: List[List[float]],
        document_metadatas: List[Dict[Any, Any]],
        collection_id: Optional[str] = None,
        collection_name: Optional[str] = None,
    ) -> Dict[Any, Any]:
        """Deletes documents from an existing collection by embedding and document metadata arrays.
        The arrays are zipped together and updates the document in the order of the two arrays.
        `column_delete()` method from [`Writer`](#writer-objects).

        Args:
            embeddings: A list of embeddings.
                Order of the embeddings should match the document_metadatas.
            document_metadatas: A list of metadata to be associated with embeddings.
                Order of these metadatas should match the embeddings.
            collection_id: The collection's id where the documents will be deleted.
                This or the `collection_name` needs to be provided.
            collection_name: The collection's name where the documents will be deleted.
                This or the `collection_id` needs to be provided.

        Returns:
            dict: delete response json

        Raises:
            ValueError: If neither collection id and collection name are provided.
            ValueError: If both collection id and collection name are provided.
            requests.exceptions.SSLError: Failure likely due to network issues.
        """
        return self.writer.column_delete(
            embeddings=embeddings,
            document_metadatas=document_metadatas,
            collection_id=collection_id,
            collection_name=collection_name,
        )

    def insert(
        self,
        documents: List[Dict[Any, Any]],
        collection_id: Optional[str] = None,
        collection_name: Optional[str] = None,
    ) -> Dict[Any, Any]:
        """Insert documents into an existing collection. `insert()` method from [`Writer`](#writer-objects).

        Args:
            documents: The documents to insert into the collection.
            collection_id: The collection's id to insert the documents to.
                This or the `collection_name` needs to be provided.
            collection_name: The collection's name to insert the documents to.
                This or the `collection_id` needs to be provided.

        Returns:
            dict: insert response json

        Raises:
            ValueError: If neither collection id and collection name are provided.
            ValueError: If both collection id and collection name are provided.
            requests.exceptions.SSLError: Failure likely due to network issues.
        """
        return self.writer.insert(
            documents=documents,
            collection_id=collection_id,
            collection_name=collection_name,
        )

    def column_insert(
        self,
        embeddings: List[List[float]],
        document_metadatas: List[Dict[Any, Any]],
        collection_id: Optional[str] = None,
        collection_name: Optional[str] = None,
    ) -> Dict[Any, Any]:
        """Insert documents into an existing collection by embedding and document metadata arrays.
        The arrays are zipped together and inserted as a document in the order of the two arrays.
        `column_insert()` method from [`Writer`](#writer-objects).

        Args:
            embeddings: A list of embeddings.
                Order of the embeddings should match the document_metadatas.
            document_metadatas: A list of metadata to be associated with embeddings.
                Order of these metadatas should match the embeddings.
            collection_id: The collection's id to insert the documents to.
                This or the `collection_name` needs to be provided.
            collection_name: The collection's name to insert the documents to.
                This or the `collection_id` needs to be provided.

        Returns:
            dict: insert response json

        Raises:
            ValueError: If neither collection id and collection name are provided.
            ValueError: If both collection id and collection name are provided.
            requests.exceptions.SSLError: Failure likely due to network issues.
        """
        return self.writer.column_insert(
            embeddings=embeddings,
            document_metadatas=document_metadatas,
            collection_id=collection_id,
            collection_name=collection_name,
        )

    def query(
        self,
        sql: Optional[str] = None,
        collection_id: Optional[str] = None,
        collection_name: Optional[str] = None,
        query_embedding: Optional[List[float]] = None,
        params: Optional[List[Any]] = None,
        text_search_query: Optional[List[str]] = None,
        text_search_weight: Optional[float] = None,
        tokenizer_type: Optional[reader.TokenizerType] = None,
    ) -> Dict[Any, Any]:
        """Queries a collection. This could be by sql or query embeddings.
        `query()` method from [`Reader`](#reader-objects).

        Args:
            sql: Raw SQL to run against the collection.
            collection_id: The collection's id where the query will happen.
                This or the `collection_name` needs to be provided.
            collection_name: The collection's name where the query will happen.
                This or the `collection_id` needs to be provided.
            query_embedding: An embedding to query against the collection using similarity search.
            params: values for parameterized sql

        Returns:
            dict: query response json

        Raises:
            ValueError: If neither collection id and collection name are provided.
            ValueError: If both collection id and collection name are provided.
            requests.exceptions.SSLError: Failure likely due to network issues.
        """
        return self.reader.query(
            sql=sql,
            collection_id=collection_id,
            collection_name=collection_name,
            query_embedding=query_embedding,
            params=params,
            text_search_query=text_search_query,
            text_search_weight=text_search_weight,
            tokenizer_type=tokenizer_type,
        )

    def infer_schema(
        self,
        collection_id: Optional[str] = None,
        collection_name: Optional[str] = None,
    ) -> Dict[Any, Any]:
        """Infers the schema of a particular collection.
        Gives the results back by column name and the inferred type for that column.
        `infer_schema()` method from [`Reader`](#reader-objects).

        Args:
            collection_id: The collection's id where the query will happen.
                This or the `collection_name` needs to be provided.
            collection_name: The collection's name where the query will happen.
                This or the `collection_id` needs to be provided.

        Returns:
            dict: infer schema response json

        Raises:
            ValueError: If neither collection id and collection name are provided.
            ValueError: If both collection id and collection name are provided.
            requests.exceptions.SSLError: Failure likely due to network issues.
        """
        return self.reader.infer_schema(
            collection_id=collection_id,
            collection_name=collection_name,
        )

    def update(
        self,
        documents: List[Dict[Any, Any]],
        collection_id: Optional[str] = None,
        collection_name: Optional[str] = None,
    ) -> Dict[Any, Any]:
        """Update documents in an existing collection. `update()` method in
        [`Writer`](#writer-objects).

        Args:
            documents: The documents to update in the collection.
            collection_id: The collection's id where the documents will be updated.
                This or the `collection_name` needs to be provided.
            collection_name: The collection's name where the documents will be updated.
                This or the `collection_id` needs to be provided.

        Returns:
            dict: update response json

        Raises:
            ValueError: If neither collection id and collection name are provided.
            ValueError: If both collection id and collection name are provided.
            requests.exceptions.SSLError: Failure likely due to network issues.
        """
        return self.writer.update(
            documents=documents,
            collection_id=collection_id,
            collection_name=collection_name,
        )

    def column_update(
        self,
        embeddings: List[List[float]],
        document_metadatas: List[Dict[Any, Any]],
        collection_id: Optional[str] = None,
        collection_name: Optional[str] = None,
    ) -> Dict[Any, Any]:
        """Updates documents for an existing collection by embedding and document metadata arrays.
        The arrays are zipped together and updates the document in the order of the two arrays.
        `column_update()` method from [`Writer`](#writer-objects).

        Args:
            embeddings: A list of embeddings.
                Order of the embeddings should match the document_metadatas.
            document_metadatas: A list of metadata to be associated with embeddings.
                Order of these metadatas should match the embeddings.
            collection_id: The collection's id where the documents will be updated.
                This or the `collection_name` needs to be provided.
            collection_name: The collection's name where the documents will be updated.
                This or the `collection_id` needs to be provided.

        Returns:
            dict: update response json

        Raises:
            ValueError: If neither collection id and collection name are provided.
            ValueError: If both collection id and collection name are provided.
            requests.exceptions.SSLError: Failure likely due to network issues.
        """
        return self.writer.column_update(
            embeddings=embeddings,
            document_metadatas=document_metadatas,
            collection_id=collection_id,
            collection_name=collection_name,
        )

    def create_collection(
        self, collection_name: str, dimensionality: int
    ) -> Dict[Any, Any]:
        """Creates a collection by name and dimensionality. Dimensionality
        should be greater than 0. `create_collection()` method from [`Writer`](#writer-objects).

        Args:
            collection_name: The name of the collection that will be created.
            dimensionality: The number of dimensions the collection will have.
                Must be an int larger than 0.

        Returns:
            dict: create collections response json

        Raises:
            ValueError: If dimensionality is 0 or less.
            requests.exceptions.SSLError: Failure likely due to network issues.
        """
        return self.writer.create_collection(
            collection_name=collection_name,
            dimensionality=dimensionality,
        )

    def delete_collection(self, collection_id: str) -> Dict[Any, Any]:
        """Deletes a collection. `delete_collection()` method from [`Writer`](#writer-objects)."""
        return self.writer.delete_collection(
            collection_id=collection_id,
        )



import logging
import more_itertools
from enum import Enum
from typing import Any, Dict, Hashable, List, Optional
from uuid import UUID

import requests

from starpoint._utils import (
    _build_header,
    _check_collection_identifier_collision,
    _validate_host,
)


LOGGER = logging.getLogger(__name__)

# Host
EMBEDDING_URL = "https://embedding.starpoint.ai"

# Endpoints
EMBED_PATH = "/api/v1/embed"

# Error and warning messages
SSL_ERROR_MSG = "Request failed due to SSLError. Error is likely due to invalid API key. Please check if your API is correct and still valid."
TEXT_METADATA_LENGTH_MISMATCH_WARNING = (
    "The length of the texts and metadatas provided are different. There may be a mismatch "
    "between texts and the metadatas length; this may cause undesired results between the joining of "
    "embeddings and metadatas."
)


class EmbeddingModel(Enum):
    MINILM = "MINI_LM"


class EmbeddingClient(object):
    """Client for the embedding endpoints."""

    def __init__(self, api_key: UUID, host: Optional[str] = None):
        if host is None:
            host = EMBEDDING_URL

        self.host = _validate_host(host)
        self.api_key = api_key

    def embed(
        self,
        texts: List[str],
        model: EmbeddingModel,
    ) -> Dict[str, List[Dict]]:
        """Takes some texts creates embeddings using a model in starpoint. This is a
        version of `embed_and_join_metadata_by_column` where joining metadata with the result is
        not necessary. The same API is used for the two methods.

        Args:
            texts: List of strings to create embeddings from.
            model: An enum choice from EmbeddingModel.

        Returns:
            dict: Result with list of texts, metadata, and embeddings.

        Raises:
            requests.exceptions.SSLError: Failure likely due to network issues.
        """
        text_embedding_items = [{"text": text, "metadata": None} for text in texts]
        return self.embed_items(text_embedding_items=text_embedding_items, model=model)

    def embed_and_join_metadata_by_columns(
        self,
        texts: List[str],
        metadatas: List[Dict],
        model: EmbeddingModel,
    ) -> Dict[str, List[Dict]]:
        """Takes some texts and creates embeddings using a model in starpoint. Prefer using `embed_and_join_metadata` or
        `embed_items` instead, as mismatched `texts` and `metadatas` will output undesirable results.
        Under the hood this is using `embed_items`.

        Args:
            texts: List of strings to create embeddings from.
            metadatas: List of metadata to join to the string and embedding when the embedding operation is complete.
                This metadata makes your embeddings queryable within starpoint.
            model: An enum choice from EmbeddingModel.

        Returns:
            dict: Result with list of texts, metadata, and embeddings.

        Raises:
            requests.exceptions.SSLError: Failure likely due to network issues.
        """
        if not isinstance(texts, list):
            raise ValueError("texts passed was not of type list")
        if not isinstance(metadatas, list):
            raise ValueError("metadatas passed was not of type list")

        if len(texts) != len(metadatas):
            LOGGER.warning(TEXT_METADATA_LENGTH_MISMATCH_WARNING)

        text_embedding_items = [
            {
                "text": text,
                "metadata": metadata,
            }
            for text, metadata in zip(texts, metadatas)
        ]
        return self.embed_items(text_embedding_items=text_embedding_items, model=model)

    def embed_and_join_metadata(
        self,
        text_embedding_items: List[Dict],
        embedding_key: Hashable,
        model: EmbeddingModel,
    ) -> Dict[str, List[Dict]]:
        """Takes some texts and creates embeddings using a model in starpoint, and joins them to
        all additional data as metadata. Under the hood this is using `embed_and_join_metadata_by_columns`
        which is using `embed_items`.

        Args:
            text_embedding_items: List of dicts of data to create embeddings from.
            embedding_key: the key in each item used to create embeddings from.
                e.g. `"context"` would be passed if each item looks like this: `{"context": "embed this text"}`
            model: An enum choice from EmbeddingModel.

        Returns:
            dict: Result with list of texts, metadata, and embeddings.

        Raises:
            requests.exceptions.SSLError: Failure likely due to network issues.
        """
        if not text_embedding_items:
            raise ValueError("text_embedding_items received an empty list.")

        texts = list(map(lambda item: item.get(embedding_key), text_embedding_items))
        if not all(texts):
            unqualified_indices = list(
                more_itertools.locate(texts, lambda x: x is None)
            )
            raise ValueError(
                "The following indices had items that did not have the "
                f"{embedding_key}:\n {unqualified_indices}"
            )

        # We can also do this operation in the first map that creates texts, but that might make additional operations
        # in here a lot more annoying. It's an optimization that shouldn't happen right now.
        list(map(lambda item: item.pop(embedding_key), text_embedding_items))

        return self.embed_and_join_metadata_by_columns(
            texts=texts, metadatas=text_embedding_items, model=model
        )

    def embed_items(
        self,
        text_embedding_items: List[Dict],
        model: EmbeddingModel,
    ) -> Dict[str, List[Dict]]:
        """Takes items with text and metadata, and embeds the text using a model in starpoint. Metadata is joined with
        the results for ergonomics.

        Args:
            text_embedding_items: List of dict where the text and metadata are paired together
            model: An enum choice from EmbeddingModel.

        Returns:
            dict: Result with list of texts, metadata, and embeddings.

        Raises:
            requests.exceptions.SSLError: Failure likely due to network issues.
        """

        request_data = dict(items=text_embedding_items, model=model.value)
        try:
            response = requests.post(
                url=f"{self.host}{EMBED_PATH}",
                json=request_data,
                headers=_build_header(
                    api_key=self.api_key,
                    additional_headers={"Content-Type": "application/json"},
                ),
            )
        except requests.exceptions.SSLError as e:
            LOGGER.error(SSL_ERROR_MSG)
            raise

        if not response.ok:
            LOGGER.error(
                f"Request failed with status code {response.status_code} "
                f"and the following message:\n{response.text}"
            )
            return {}
        return response.json()



from typing import List
from uuid import uuid4
from unittest.mock import MagicMock, patch

import pandas as pd
import pytest

from starpoint import pandas


def test__check_column_lenth_too_few_columns():
    too_few_column_dataframe = pd.DataFrame([])

    with pytest.raises(ValueError, match=pandas.TOO_FEW_COLUMN_ERROR):
        pandas._check_column_length(too_few_column_dataframe)


def test__get_aggregate_column_values_from_dataframe_exclude_partial():
    test_dataframe = pd.DataFrame([[1, 2], [3, 4]], columns=["include", "exclude"])

    expected_list = [{"include": 1}, {"include": 3}]

    actual_list = pandas._get_aggregate_column_values_from_dataframe(
        test_dataframe, ["exclude"]
    )

    assert actual_list == expected_list


def test__get_aggregate_column_values_from_dataframe_exclude_all():
    test_dataframe = pd.DataFrame([[1, 2], [3, 4]], columns=["exclude1", "exclude2"])

    actual_list = pandas._get_aggregate_column_values_from_dataframe(
        test_dataframe, ["exclude1", "exclude2"]
    )

    assert actual_list == []


@pytest.mark.parametrize("exclude_columns", [["random col"], ["exclude", "random col"]])
def test__get_aggregate_column_values_from_dataframe_exclude_unrelated(exclude_columns):
    test_dataframe = pd.DataFrame([[1, 2], [3, 4]], columns=["include", "exclude"])

    with pytest.raises(ValueError):
        pandas._get_aggregate_column_values_from_dataframe(
            test_dataframe, exclude_columns
        )


def test_init_pandas_client():
    mock_startpoint_client = MagicMock()
    pandas_client = pandas.PandasClient(mock_startpoint_client)

    assert pandas_client.starpoint == mock_startpoint_client


@patch("starpoint.pandas._check_column_length")
def test_insert_by_dataframe_success(check_column_mock: MagicMock):
    """Tests a successful insertion operation."""
    mock_startpoint_client = MagicMock()
    pandas_client = pandas.PandasClient(mock_startpoint_client)

    test_dataframe = pd.DataFrame(
        [[1, 2]],
        columns=["embedding", "metadata"],
    )

    pandas_client.insert_by_dataframe(test_dataframe)

    check_column_mock.assert_called_once_with(test_dataframe)
    mock_startpoint_client.column_insert.assert_called_once_with(
        embeddings=[1],
        document_metadatas=[{"metadata": 2}],
        collection_id=None,
        collection_name=None,
    )


def test_insert_by_dataframe_missing_embedding_column():
    mock_startpoint_client = MagicMock()
    pandas_client = pandas.PandasClient(mock_startpoint_client)

    missing_embedding_column_dataframe = pd.DataFrame(
        [[1, 2]],
        columns=["metadata", "extra"],
    )

    with pytest.raises(KeyError) as excinfo:
        pandas_client.insert_by_dataframe(missing_embedding_column_dataframe)

    assert (
        pandas.MISSING_COLUMN.substitute(column_name="embedding")
        in excinfo.value.__notes__
    )


@patch("starpoint.pandas._check_column_length")
def test_update_by_dataframe_success(check_column_mock: MagicMock):
    """Tests a successful update operation."""
    mock_startpoint_client = MagicMock()
    pandas_client = pandas.PandasClient(mock_startpoint_client)

    test_dataframe = pd.DataFrame(
        [[1, 2]],
        columns=["embedding", "metadata"],
    )

    pandas_client.update_by_dataframe(test_dataframe)

    check_column_mock.assert_called_once_with(test_dataframe)
    mock_startpoint_client.column_update.assert_called_once_with(
        embeddings=[1],
        document_metadatas=[{"metadata": 2}],
        collection_id=None,
        collection_name=None,
    )


def test_update_by_dataframe_missing_embedding_column():
    mock_startpoint_client = MagicMock()
    pandas_client = pandas.PandasClient(mock_startpoint_client)

    missing_embedding_column_dataframe = pd.DataFrame(
        [[1, 2]],
        columns=["metadata", "extra"],
    )

    with pytest.raises(KeyError) as excinfo:
        pandas_client.update_by_dataframe(missing_embedding_column_dataframe)

    assert (
        pandas.MISSING_COLUMN.substitute(column_name="embedding")
        in excinfo.value.__notes__
    )


@patch("starpoint.pandas._check_column_length")
def test_delete_by_dataframe_success(check_column_mock: MagicMock):
    """Tests a successful delete operation."""
    mock_startpoint_client = MagicMock()
    pandas_client = pandas.PandasClient(mock_startpoint_client)

    test_dataframe = pd.DataFrame(
        [[1, 2]],
        columns=["embedding", "metadata"],
    )

    pandas_client.delete_by_dataframe(test_dataframe)

    check_column_mock.assert_called_once_with(test_dataframe)
    mock_startpoint_client.column_delete.assert_called_once_with(
        embeddings=[1],
        document_metadatas=[{"metadata": 2}],
        collection_id=None,
        collection_name=None,
    )


def test_delete_by_dataframe_missing_embedding_column():
    mock_startpoint_client = MagicMock()
    pandas_client = pandas.PandasClient(mock_startpoint_client)

    missing_embedding_column_dataframe = pd.DataFrame(
        [[1, 2]],
        columns=["metadata", "extra"],
    )

    with pytest.raises(KeyError) as excinfo:
        pandas_client.delete_by_dataframe(missing_embedding_column_dataframe)

    assert (
        pandas.MISSING_COLUMN.substitute(column_name="embedding")
        in excinfo.value.__notes__
    )



import logging
from typing import Dict, Optional
from uuid import UUID

import requests
import validators

LOGGER = logging.getLogger(__name__)

# Key in the header value
API_HEADER_KEY = "x-starpoint-key"

# Health check for the hosts
HEALTH_CHECK_MESSAGE = "hello."

# Error messages for collections
NO_COLLECTION_VALUE_ERROR = (
    "Please provide at least one value for either collection_id or collection_name."
)
MULTI_COLLECTION_VALUE_ERROR = (
    "Please only provide either collection_id or collection_name in your request."
)
NO_HOST_ERROR = "No host value provided. A host must be provided."


def _build_header(api_key: UUID, additional_headers: Optional[Dict[str, str]] = None):
    """Builds the header for starpoint requests.
    Args:
        api_key: API key for starpoint.
        additional_headers: additional headers that need to be part of the header;
            for example `Content-Type`.

    Returns:
        dict: Header for the request.
    """
    header = {API_HEADER_KEY: str(api_key)}
    if additional_headers is not None:
        header.update(additional_headers)
    return header


def _check_host_health(hostname: str):
    """Check that the host is healthy. This is non-blocking as long as the host responds with a 200.

    Args:
        hostname: hostname of the service.

    Raises:
        AssertionError: host does not return a 200 when a GET request is sent.
    """
    resp = requests.get(hostname)
    assert (
        resp.ok
    ), f"Host cannot be validated, response from host {hostname}: {resp.text}"
    if resp.text != HEALTH_CHECK_MESSAGE:
        LOGGER.warning(
            f"{hostname} returned {resp.text} instead of {HEALTH_CHECK_MESSAGE}; host may be unhealthy and "
            "may be unable to serve requests."
        )


def _validate_host(host: str):
    """Check the hostname and health of the host before returning a properly formatted hostname.
    Args:
        host: hostname of the service.

    Returns:
        str: hostname, trimmed of extra backslashes if necessary.

    Raises:
        ValueError: Empty string is provided for host.
        ValueError: Invalid url format.
    """
    if not host:
        raise ValueError(NO_HOST_ERROR)
    elif validators.url(host) is not True:  # type: ignore
        raise ValueError(f"Provided host {host} is not a valid url format.")

    # Make sure we don't have dangling backslashes in the url during url composition later
    trimmed_hostname = host.rstrip("/")

    _check_host_health(trimmed_hostname)

    return trimmed_hostname


def _check_collection_identifier_collision(
    collection_id: Optional[str] = None, collection_name: Optional[str] = None
):
    """Check whether both or neither collection_id and collection_name is provided.
    Args:
        collection_id: The collection's id where the query will happen.
            This or the `collection_name` needs to be provided.
        collection_name: The collection's name where the query will happen.
            This or the `collection_id` needs to be provided.

    Raises:
        ValueError: If neither collection id and collection name are provided.
        ValueError: If both collection id and collection name are provided.
    """
    if collection_id is None and collection_name is None:
        raise ValueError(NO_COLLECTION_VALUE_ERROR)
    elif collection_id and collection_name:
        raise ValueError(MULTI_COLLECTION_VALUE_ERROR)



import logging
from string import Template
from typing import Any, Dict, List, Optional

import pandas as pd

from starpoint import db

LOGGER = logging.getLogger(__name__)

EMBEDDING_COLUMN_NAME = "embedding"

TOO_FEW_COLUMN_ERROR = """Not enough columns in dataframe provided. Please make sure to provide a
column for at least embeddings. For examples of what this should look like visit:
https://docs.starpoint.ai/"""
MISSING_COLUMN = Template(
    'Missing column name "$column_name" expected for starpoint write operations.'
)


def _check_column_length(dataframe: pd.DataFrame):
    """Checks that, by length, we satisfy basic requirements for a starpoint write operation.
    For any starpoint write operations, we need at least the embeddings column.
    """
    if len(dataframe.columns) < 1:
        raise ValueError(TOO_FEW_COLUMN_ERROR)


def _get_aggregate_column_values_from_dataframe(
    dataframe: pd.DataFrame, exclude_column_names: List[str]
) -> List[Dict]:
    """Gets a dataframe of everything except for the "embedding" column then produce
    a list of row-wise dicts that will be loaded as the metadata. For example:

    df = DataFrame([[1,2,3], [4,5,6]], columns=["embedding","b","c"]
    metadata_column_values will be [{'b': 2, 'c': 3}, {'b': 5, 'c': 6}]
    """
    if not all((True if name in dataframe else False for name in exclude_column_names)):
        raise ValueError(
            f"{exclude_column_names} has values that does not exist in "
            "the dataframe and cannot be excluded."
        )

    metadatas_df = dataframe.loc[:, ~dataframe.columns.isin(exclude_column_names)]
    return metadatas_df.to_dict(orient="records")


def _get_column_value_from_dataframe(dataframe: pd.DataFrame, column_name: str) -> List:
    try:
        column = dataframe[column_name]
    except KeyError as e:
        e.add_note(MISSING_COLUMN.substitute(column_name=column_name))
        raise
    # TODO: check values using df to make sure values aren't totally bogus
    return column.values.tolist()


class PandasClient(object):
    def __init__(
        self,
        starpoint: db.Client,
    ):
        self.starpoint = starpoint

    def insert_by_dataframe(
        self,
        dataframe: pd.DataFrame,
        collection_id: Optional[str] = None,
        collection_name: Optional[str] = None,
        embedding_column_name: str = EMBEDDING_COLUMN_NAME,
    ) -> Dict[Any, Any]:
        _check_column_length(dataframe)
        embedding_column_values = _get_column_value_from_dataframe(
            dataframe,
            embedding_column_name,
        )

        metadata_column_values = _get_aggregate_column_values_from_dataframe(
            dataframe,
            [embedding_column_name],
        )

        self.starpoint.column_insert(
            embeddings=embedding_column_values,
            document_metadatas=metadata_column_values,
            collection_id=collection_id,
            collection_name=collection_name,
        )

    def update_by_dataframe(
        self,
        dataframe: pd.DataFrame,
        collection_id: Optional[str] = None,
        collection_name: Optional[str] = None,
        embedding_column_name: str = EMBEDDING_COLUMN_NAME,
    ) -> Dict[Any, Any]:
        _check_column_length(dataframe)
        embedding_column_values = _get_column_value_from_dataframe(
            dataframe,
            embedding_column_name,
        )

        metadata_column_values = _get_aggregate_column_values_from_dataframe(
            dataframe,
            [embedding_column_name],
        )

        self.starpoint.column_update(
            embeddings=embedding_column_values,
            document_metadatas=metadata_column_values,
            collection_id=collection_id,
            collection_name=collection_name,
        )

    def delete_by_dataframe(
        self,
        dataframe: pd.DataFrame,
        collection_id: Optional[str] = None,
        collection_name: Optional[str] = None,
        embedding_column_name: str = EMBEDDING_COLUMN_NAME,
    ) -> Dict[Any, Any]:
        _check_column_length(dataframe)
        embedding_column_values = _get_column_value_from_dataframe(
            dataframe,
            embedding_column_name,
        )

        metadata_column_values = _get_aggregate_column_values_from_dataframe(
            dataframe,
            [embedding_column_name],
        )

        self.starpoint.column_delete(
            embeddings=embedding_column_values,
            document_metadatas=metadata_column_values,
            collection_id=collection_id,
            collection_name=collection_name,
        )



from uuid import UUID, uuid4
from unittest.mock import MagicMock, patch

import pytest
from _pytest.monkeypatch import MonkeyPatch
from requests.exceptions import SSLError

from starpoint import writer


@pytest.fixture(scope="session")
def api_uuid() -> UUID:
    return uuid4()


@pytest.fixture(scope="session")
@patch("starpoint._utils._check_host_health")
def mock_writer(host_health_mock: MagicMock, api_uuid: UUID) -> writer.Writer:
    return writer.Writer(api_uuid)


def test_writer_default_init(mock_writer: writer.Writer, api_uuid: UUID):
    assert mock_writer.host
    assert mock_writer.host == writer.WRITER_URL
    assert mock_writer.api_key == api_uuid


def test_writer_init_non_default_host(api_uuid: UUID):
    test_host = "http://www.example.com"
    test_writer = writer.Writer(api_key=api_uuid, host=test_host)

    assert test_writer.host
    assert test_writer.host == test_host
    assert test_writer.api_key == api_uuid


@patch("starpoint.writer._check_collection_identifier_collision")
@patch("starpoint.writer.requests")
def test_writer_delete_by_collection_id(
    requests_mock: MagicMock, collision_mock: MagicMock, mock_writer: writer.Writer
):
    test_uuid = uuid4()

    mock_writer.delete(documents=[uuid4()], collection_id=test_uuid)

    collision_mock.assert_called_once_with(test_uuid, None)
    requests_mock.delete.assert_called_once()


@patch("starpoint.writer._check_collection_identifier_collision")
@patch("starpoint.writer.requests")
def test_writer_delete_by_collection_name(
    requests_mock: MagicMock, collision_mock: MagicMock, mock_writer: writer.Writer
):
    test_collection_name = "mock_collection_name"

    mock_writer.delete(documents=[uuid4()], collection_name=test_collection_name)

    collision_mock.assert_called_once_with(None, test_collection_name)
    requests_mock.delete.assert_called_once()


@patch("starpoint.writer.requests")
def test_writer_delete_not_200(
    requests_mock: MagicMock, mock_writer: writer.Writer, monkeypatch: MonkeyPatch
):
    requests_mock.delete().ok = False

    expected_json = {}

    logger_mock = MagicMock()
    monkeypatch.setattr(writer, "LOGGER", logger_mock)

    actual_json = mock_writer.delete(
        documents=[uuid4()], collection_name="mock_collection_name"
    )

    requests_mock.delete.assert_called()
    logger_mock.error.assert_called_once()
    assert actual_json == expected_json


@patch("starpoint.writer.requests")
def test_writer_delete_SSLError(
    requests_mock: MagicMock, mock_writer: writer.Writer, monkeypatch: MonkeyPatch
):
    requests_mock.exceptions.SSLError = SSLError
    requests_mock.delete.side_effect = SSLError("mock exception")

    logger_mock = MagicMock()
    monkeypatch.setattr(writer, "LOGGER", logger_mock)

    with pytest.raises(SSLError, match="mock exception"):
        mock_writer.delete(documents=[uuid4()], collection_name="mock_collection_name")

    logger_mock.error.assert_called_once_with(writer.SSL_ERROR_MSG)


@patch("starpoint.writer._check_collection_identifier_collision")
@patch("starpoint.writer.requests")
def test_writer_insert_by_collection_id(
    requests_mock: MagicMock, collision_mock: MagicMock, mock_writer: writer.Writer
):
    test_uuid = uuid4()

    mock_writer.insert(documents=[uuid4()], collection_id=test_uuid)

    collision_mock.assert_called_once_with(test_uuid, None)
    requests_mock.post.assert_called_once()


@patch("starpoint.writer._check_collection_identifier_collision")
@patch("starpoint.writer.requests")
def test_writer_insert_by_collection_name(
    requests_mock: MagicMock, collision_mock: MagicMock, mock_writer: writer.Writer
):
    test_collection_name = "mock_collection_name"

    mock_writer.insert(documents=[uuid4()], collection_name=test_collection_name)

    collision_mock.assert_called_once_with(None, test_collection_name)
    requests_mock.post.assert_called_once()


@patch("starpoint.writer.requests")
def test_writer_insert_not_200(
    requests_mock: MagicMock, mock_writer: writer.Writer, monkeypatch: MonkeyPatch
):
    requests_mock.post().ok = False

    expected_json = {}

    logger_mock = MagicMock()
    monkeypatch.setattr(writer, "LOGGER", logger_mock)

    actual_json = mock_writer.insert(
        documents=[uuid4()], collection_name="mock_collection_name"
    )

    requests_mock.post.assert_called()
    logger_mock.error.assert_called_once()
    assert actual_json == expected_json


@patch("starpoint.writer.requests")
def test_writer_insert_SSLError(
    requests_mock: MagicMock, mock_writer: writer.Writer, monkeypatch: MonkeyPatch
):
    requests_mock.exceptions.SSLError = SSLError
    requests_mock.post.side_effect = SSLError("mock exception")

    logger_mock = MagicMock()
    monkeypatch.setattr(writer, "LOGGER", logger_mock)

    with pytest.raises(SSLError, match="mock exception"):
        mock_writer.insert(documents=[uuid4()], collection_name="mock_collection_name")

    logger_mock.error.assert_called_once_with(writer.SSL_ERROR_MSG)


@patch("starpoint.writer.Writer.insert")
def test_writer_column_insert(insert_mock: MagicMock, mock_writer: writer.Writer):
    test_embeddings = [0.88, 0.71]
    test_document_metadatas = [{"mock": "metadata"}, {"mock2": "metadata2"}]
    expected_insert_document = [
        {
            "embedding": test_embeddings[0],
            "metadata": test_document_metadatas[0],
        },
        {
            "embedding": test_embeddings[1],
            "metadata": test_document_metadatas[1],
        },
    ]

    mock_writer.column_insert(
        embeddings=test_embeddings, document_metadatas=test_document_metadatas
    )

    insert_mock.assert_called_once_with(
        documents=expected_insert_document,
        collection_id=None,
        collection_name=None,
    )


@patch("starpoint.writer.Writer.insert")
def test_writer_column_insert_collection_id_collection_name_passed_through(
    insert_mock: MagicMock, mock_writer: writer.Writer
):
    test_embeddings = [0.88]
    test_document_metadatas = [{"mock": "metadata"}]
    expected_insert_document = [
        {
            "embedding": test_embeddings[0],
            "metadata": test_document_metadatas[0],
        },
    ]
    expected_collection_id = "mock_id"
    expected_collection_name = "mock_name"

    mock_writer.column_insert(
        embeddings=test_embeddings,
        document_metadatas=test_document_metadatas,
        collection_id=expected_collection_id,
        collection_name=expected_collection_name,
    )

    insert_mock.assert_called_once_with(
        documents=expected_insert_document,
        collection_id=expected_collection_id,
        collection_name=expected_collection_name,
    )


@patch("starpoint.writer.Writer.insert")
def test_writer_column_insert_shorter_metadatas_length(
    insert_mock: MagicMock, mock_writer: writer.Writer, monkeypatch: MonkeyPatch
):
    test_embeddings = [0.88, 0.71]
    test_document_metadatas = [{"mock": "metadata"}]
    expected_insert_document = [
        {
            "embedding": test_embeddings[0],
            "metadata": test_document_metadatas[0],
        },
    ]

    logger_mock = MagicMock()
    monkeypatch.setattr(writer, "LOGGER", logger_mock)

    mock_writer.column_insert(
        embeddings=test_embeddings, document_metadatas=test_document_metadatas
    )

    logger_mock.warning.assert_called_once_with(
        writer.EMBEDDING_METADATA_LENGTH_MISMATCH_WARNING
    )
    insert_mock.assert_called_once_with(
        documents=expected_insert_document,
        collection_id=None,
        collection_name=None,
    )


@patch("starpoint.writer.Writer.insert")
def test_writer_column_insert_shorter_embeddings_length(
    insert_mock: MagicMock, mock_writer: writer.Writer, monkeypatch: MonkeyPatch
):
    test_embeddings = [0.88]
    test_document_metadatas = [{"mock": "metadata"}, {"mock2": "metadata2"}]
    expected_insert_document = [
        {
            "embedding": test_embeddings[0],
            "metadata": test_document_metadatas[0],
        },
    ]

    logger_mock = MagicMock()
    monkeypatch.setattr(writer, "LOGGER", logger_mock)

    mock_writer.column_insert(
        embeddings=test_embeddings, document_metadatas=test_document_metadatas
    )

    logger_mock.warning.assert_called_once_with(
        writer.EMBEDDING_METADATA_LENGTH_MISMATCH_WARNING
    )
    insert_mock.assert_called_once_with(
        documents=expected_insert_document,
        collection_id=None,
        collection_name=None,
    )


@patch("starpoint.writer._check_collection_identifier_collision")
@patch("starpoint.writer.requests")
def test_writer_update_by_collection_id(
    requests_mock: MagicMock, collision_mock: MagicMock, mock_writer: writer.Writer
):
    test_uuid = uuid4()

    mock_writer.update(documents=[uuid4()], collection_id=test_uuid)

    collision_mock.assert_called_once_with(test_uuid, None)
    requests_mock.patch.assert_called_once()


@patch("starpoint.writer._check_collection_identifier_collision")
@patch("starpoint.writer.requests")
def test_writer_update_by_collection_name(
    requests_mock: MagicMock, collision_mock: MagicMock, mock_writer: writer.Writer
):
    test_collection_name = "mock_collection_name"

    mock_writer.update(documents=[uuid4()], collection_name=test_collection_name)

    collision_mock.assert_called_once_with(None, test_collection_name)
    requests_mock.patch.assert_called_once()


@patch("starpoint.writer.requests")
def test_writer_update_not_200(
    requests_mock: MagicMock, mock_writer: writer.Writer, monkeypatch: MonkeyPatch
):
    requests_mock.patch().ok = False

    expected_json = {}

    logger_mock = MagicMock()
    monkeypatch.setattr(writer, "LOGGER", logger_mock)

    actual_json = mock_writer.update(
        documents=[uuid4()], collection_name="mock_collection_name"
    )

    requests_mock.patch.assert_called()
    logger_mock.error.assert_called_once()
    assert actual_json == expected_json


@patch("starpoint.writer.requests")
def test_writer_update_SSLError(
    requests_mock: MagicMock, mock_writer: writer.Writer, monkeypatch: MonkeyPatch
):
    requests_mock.exceptions.SSLError = SSLError
    requests_mock.patch.side_effect = SSLError("mock exception")

    logger_mock = MagicMock()
    monkeypatch.setattr(writer, "LOGGER", logger_mock)

    with pytest.raises(SSLError, match="mock exception"):
        mock_writer.update(documents=[uuid4()], collection_name="mock_collection_name")

    logger_mock.error.assert_called_once_with(writer.SSL_ERROR_MSG)


@patch("starpoint.writer.Writer.update")
def test_writer_column_update(update_mock: MagicMock, mock_writer: writer.Writer):
    test_embeddings = [0.88, 0.71]
    test_document_metadatas = [{"mock": "metadata"}, {"mock2": "metadata2"}]
    expected_update_document = [
        {
            "embedding": test_embeddings[0],
            "metadata": test_document_metadatas[0],
        },
        {
            "embedding": test_embeddings[1],
            "metadata": test_document_metadatas[1],
        },
    ]

    mock_writer.column_update(
        embeddings=test_embeddings, document_metadatas=test_document_metadatas
    )

    update_mock.assert_called_once_with(
        documents=expected_update_document,
        collection_id=None,
        collection_name=None,
    )


@patch("starpoint.writer.Writer.update")
def test_writer_column_update_collection_id_collection_name_passed_through(
    update_mock: MagicMock, mock_writer: writer.Writer
):
    test_embeddings = [0.88]
    test_document_metadatas = [{"mock": "metadata"}]
    expected_update_document = [
        {
            "embedding": test_embeddings[0],
            "metadata": test_document_metadatas[0],
        },
    ]
    expected_collection_id = "mock_id"
    expected_collection_name = "mock_name"

    mock_writer.column_update(
        embeddings=test_embeddings,
        document_metadatas=test_document_metadatas,
        collection_id=expected_collection_id,
        collection_name=expected_collection_name,
    )

    update_mock.assert_called_once_with(
        documents=expected_update_document,
        collection_id=expected_collection_id,
        collection_name=expected_collection_name,
    )


@patch("starpoint.writer.Writer.update")
def test_writer_column_insert_shorter_metadatas_length(
    update_mock: MagicMock, mock_writer: writer.Writer, monkeypatch: MonkeyPatch
):
    test_embeddings = [0.88, 0.71]
    test_document_metadatas = [{"mock": "metadata"}]
    expected_update_document = [
        {
            "embedding": test_embeddings[0],
            "metadata": test_document_metadatas[0],
        },
    ]

    logger_mock = MagicMock()
    monkeypatch.setattr(writer, "LOGGER", logger_mock)

    mock_writer.column_update(
        embeddings=test_embeddings, document_metadatas=test_document_metadatas
    )

    logger_mock.warning.assert_called_once_with(
        writer.EMBEDDING_METADATA_LENGTH_MISMATCH_WARNING
    )
    update_mock.assert_called_once_with(
        documents=expected_update_document,
        collection_id=None,
        collection_name=None,
    )


@patch("starpoint.writer.Writer.update")
def test_writer_column_update_shorter_embeddings_length(
    update_mock: MagicMock, mock_writer: writer.Writer, monkeypatch: MonkeyPatch
):
    test_embeddings = [0.88]
    test_document_metadatas = [{"mock": "metadata"}, {"mock2": "metadata2"}]
    expected_update_document = [
        {
            "embedding": test_embeddings[0],
            "metadata": test_document_metadatas[0],
        },
    ]

    logger_mock = MagicMock()
    monkeypatch.setattr(writer, "LOGGER", logger_mock)

    mock_writer.column_update(
        embeddings=test_embeddings, document_metadatas=test_document_metadatas
    )

    logger_mock.warning.assert_called_once_with(
        writer.EMBEDDING_METADATA_LENGTH_MISMATCH_WARNING
    )
    update_mock.assert_called_once_with(
        documents=expected_update_document,
        collection_id=None,
        collection_name=None,
    )


@patch("starpoint.writer.Writer.delete")
def test_writer_column_delete(delete_mock: MagicMock, mock_writer: writer.Writer):
    test_embeddings = [0.88, 0.71]
    test_document_metadatas = [{"mock": "metadata"}, {"mock2": "metadata2"}]
    expected_delete_document = [
        {
            "embedding": test_embeddings[0],
            "metadata": test_document_metadatas[0],
        },
        {
            "embedding": test_embeddings[1],
            "metadata": test_document_metadatas[1],
        },
    ]

    mock_writer.column_delete(
        embeddings=test_embeddings, document_metadatas=test_document_metadatas
    )

    delete_mock.assert_called_once_with(
        documents=expected_delete_document,
        collection_id=None,
        collection_name=None,
    )


@patch("starpoint.writer.Writer.delete")
def test_writer_column_delete_collection_id_collection_name_passed_through(
    delete_mock: MagicMock, mock_writer: writer.Writer
):
    test_embeddings = [0.88]
    test_document_metadatas = [{"mock": "metadata"}]
    expected_delete_document = [
        {
            "embedding": test_embeddings[0],
            "metadata": test_document_metadatas[0],
        },
    ]
    expected_collection_id = "mock_id"
    expected_collection_name = "mock_name"

    mock_writer.column_delete(
        embeddings=test_embeddings,
        document_metadatas=test_document_metadatas,
        collection_id=expected_collection_id,
        collection_name=expected_collection_name,
    )

    delete_mock.assert_called_once_with(
        documents=expected_delete_document,
        collection_id=expected_collection_id,
        collection_name=expected_collection_name,
    )


@patch("starpoint.writer.Writer.delete")
def test_writer_column_insert_shorter_metadatas_length(
    delete_mock: MagicMock, mock_writer: writer.Writer, monkeypatch: MonkeyPatch
):
    test_embeddings = [0.88, 0.71]
    test_document_metadatas = [{"mock": "metadata"}]
    expected_delete_document = [
        {
            "embedding": test_embeddings[0],
            "metadata": test_document_metadatas[0],
        },
    ]

    logger_mock = MagicMock()
    monkeypatch.setattr(writer, "LOGGER", logger_mock)

    mock_writer.column_delete(
        embeddings=test_embeddings, document_metadatas=test_document_metadatas
    )

    logger_mock.warning.assert_called_once_with(
        writer.EMBEDDING_METADATA_LENGTH_MISMATCH_WARNING
    )
    delete_mock.assert_called_once_with(
        documents=expected_delete_document,
        collection_id=None,
        collection_name=None,
    )


@patch("starpoint.writer.Writer.delete")
def test_writer_column_delete_shorter_embeddings_length(
    delete_mock: MagicMock, mock_writer: writer.Writer, monkeypatch: MonkeyPatch
):
    test_embeddings = [0.88]
    test_document_metadatas = [{"mock": "metadata"}, {"mock2": "metadata2"}]
    expected_delete_document = [
        {
            "embedding": test_embeddings[0],
            "metadata": test_document_metadatas[0],
        },
    ]

    logger_mock = MagicMock()
    monkeypatch.setattr(writer, "LOGGER", logger_mock)

    mock_writer.column_delete(
        embeddings=test_embeddings, document_metadatas=test_document_metadatas
    )

    logger_mock.warning.assert_called_once_with(
        writer.EMBEDDING_METADATA_LENGTH_MISMATCH_WARNING
    )
    delete_mock.assert_called_once_with(
        documents=expected_delete_document,
        collection_id=None,
        collection_name=None,
    )
